@@ -1,4 +1,4 @@
-/*  $Id: gui.c 7317 2009-05-09 08:43:04Z enrico $
+/*  $Id$
  *
  *  Copyright 2006-2009 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
  *
@@ -26,15 +26,18 @@
 #endif
 
 #include <string.h>
+#include <stdarg.h>
+
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
-#include <libxfcegui4/libxfcegui4.h>
+#include <libxfce4util/libxfce4util.h>
 
 #include "common.h"
 #include "gui.h"
 #include "sexy-icon-entry.h"
 #include "searchentry.h"
 #include "inline-icon.h"
+#include "speedreader.h"
 
 
 
@@ -42,7 +45,6 @@ static gboolean hovering_over_link = FALSE;
 static GdkCursor *hand_cursor = NULL;
 static GdkCursor *regular_cursor = NULL;
 static gboolean entry_is_dirty = FALSE;
-static const GdkColor error_color = { 0, 0x8000, 0, 0 }; /* dark red */
 
 
 /* all textview_* functions are from the gtk-demo app to get links in the textview working */
@@ -69,6 +71,9 @@ static void textview_follow_if_link(GtkWidget *text_view, GtkTextIter *iter, Dic
 			gboolean browser_started = dict_start_web_query(dd, dd->searched_word);
 			if (browser_started && dd->is_plugin)
 				gtk_widget_hide(dd->window);
+
+			gdk_window_set_cursor(gtk_text_view_get_window(
+					GTK_TEXT_VIEW(text_view), GTK_TEXT_WINDOW_TEXT), regular_cursor);
 			g_free(found_link);
 			break;
 		}
@@ -156,6 +161,7 @@ static void textview_set_cursor_if_appropriate(GtkTextView *view, gint x, gint y
 			g_free(name);
 			break;
 		}
+		g_free(name);
 	}
 
 	if (hovering != hovering_over_link)
@@ -182,8 +188,6 @@ static gboolean textview_motion_notify_event(GtkWidget *text_view, GdkEventMotio
 
 	textview_set_cursor_if_appropriate(GTK_TEXT_VIEW(text_view), x, y, event->window);
 
-	gdk_window_get_pointer(text_view->window, NULL, NULL, NULL);
-
 	return FALSE;
 }
 
@@ -270,6 +274,81 @@ static gboolean textview_button_press_cb(GtkTextView *view, GdkEventButton *even
 }
 
 
+#if GTK_CHECK_VERSION(2, 12, 0)
+static gboolean textview_query_tooltip_cb(GtkWidget *widget, gint x, gint y, gboolean keyboard_mode,
+										  GtkTooltip *tooltip, DictData *dd)
+{
+	gint tx, ty;
+	GSList *tags = NULL, *tagp = NULL;
+	GtkTextIter iter;
+
+	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget),
+		GTK_TEXT_WINDOW_WIDGET, x, y, &tx, &ty);
+
+	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(widget), &iter, tx, ty);
+
+	tags = gtk_text_iter_get_tags(&iter);
+	for (tagp = tags;  tagp != NULL;  tagp = tagp->next)
+	{
+		GtkTextTag *tag = tagp->data;
+		gchar *name;
+
+		g_object_get(G_OBJECT(tag), "name", &name, NULL);
+		if (name != NULL && strcmp("link", name) == 0)
+		{
+			gchar *target_uri = dict_get_web_query_uri(dd, dd->searched_word);
+			gtk_tooltip_set_markup(tooltip, target_uri);
+			g_free(name);
+			g_free(target_uri);
+			return TRUE;
+		}
+		g_free(name);
+	}
+	return FALSE;
+}
+#endif
+
+
+static void textview_apply_or_remove_tags(GtkTextBuffer *buffer, const gchar *tag,
+										  GtkTextIter *start, GtkTextIter *end)
+{
+	g_return_if_fail(tag != NULL);
+
+	if (*tag == '\0')
+		gtk_text_buffer_remove_all_tags(buffer, start, end);
+	else
+		gtk_text_buffer_apply_tag_by_name(buffer, tag, start, end);
+}
+
+
+/* An empty string as tag name will clear all tags on the given word */
+void dict_gui_textview_apply_tag_to_word(GtkTextBuffer *buffer, const gchar *word,
+										 GtkTextIter *pos, const gchar *first_tag, ...)
+{
+	GtkTextIter start, end;
+	gchar *s;
+
+	g_return_if_fail(word != NULL);
+	g_return_if_fail(first_tag != NULL);
+
+	/* This is a bit ugly but works:
+	 * we search the passed word backwards from 'pos' and apply the tag to it. */
+	if (gtk_text_iter_backward_search(pos, word, GTK_TEXT_SEARCH_TEXT_ONLY, &start, &end, NULL))
+	{
+		va_list args;
+
+		textview_apply_or_remove_tags(buffer, first_tag, &start, &end);
+
+		va_start(args, first_tag);
+		for (; s = va_arg(args, gchar*), s != NULL;)
+		{
+			textview_apply_or_remove_tags(buffer, s, &start, &end);
+		}
+		va_end(args);
+	}
+}
+
+
 void dict_gui_set_panel_entry_text(DictData *dd, const gchar *text)
 {
 	if (dd->panel_entry != NULL)
@@ -392,7 +471,7 @@ static void update_search_button(DictData *dd, GtkWidget *box)
 
 	if (button == NULL)
 	{
-		button = gtk_button_new_from_stock("gtk-find");
+		button = gtk_button_new_from_stock(GTK_STOCK_FIND);
 		gtk_widget_show(button);
 		gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
 		g_signal_connect(button, "clicked", G_CALLBACK(entry_button_clicked_cb), dd);
@@ -405,7 +484,7 @@ static void update_search_button(DictData *dd, GtkWidget *box)
 	{
 		case DICTMODE_DICT:
 		{
-			image = gtk_image_new_from_stock("gtk-find", GTK_ICON_SIZE_BUTTON);
+			image = gtk_image_new_from_stock(GTK_STOCK_FIND, GTK_ICON_SIZE_BUTTON);
 			break;
 		}
 		case DICTMODE_WEB:
@@ -415,7 +494,7 @@ static void update_search_button(DictData *dd, GtkWidget *box)
 		}
 		case DICTMODE_SPELL:
 		{
-			image = gtk_image_new_from_stock("gtk-spell-check", GTK_ICON_SIZE_BUTTON);
+			image = gtk_image_new_from_stock(GTK_STOCK_SPELL_CHECK, GTK_ICON_SIZE_BUTTON);
 			break;
 		}
 		default:
@@ -467,7 +546,14 @@ const guint8 *dict_gui_get_icon_data(void)
 }
 
 
- static GtkWidget *create_file_menu(DictData *dd)
+static void speedreader_clicked_cb(GtkButton *button, DictData *dd)
+{
+	GtkWidget *dialog = xfd_speed_reader_new(GTK_WINDOW(dd->window), dd);
+	gtk_widget_show(dialog);
+}
+
+
+static GtkWidget *create_file_menu(DictData *dd)
 {
 	GtkWidget *menubar, *file, *file_menu, *help, *help_menu, *menu_item;
 	GtkAccelGroup *accel_group;
@@ -482,6 +568,16 @@ const guint8 *dict_gui_get_icon_data(void)
 	file_menu = gtk_menu_new();
 	gtk_menu_item_set_submenu(GTK_MENU_ITEM(file), file_menu);
 
+	menu_item = gtk_image_menu_item_new_with_mnemonic(_("Speed _Reader"));
+	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item),
+		gtk_image_new_from_stock(GTK_STOCK_JUSTIFY_CENTER, GTK_ICON_SIZE_MENU));
+	gtk_widget_add_accelerator(menu_item, "activate", accel_group,
+			GDK_r, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
+	g_signal_connect(menu_item, "activate", G_CALLBACK(speedreader_clicked_cb), dd);
+	gtk_container_add(GTK_CONTAINER(file_menu), menu_item);
+
+	gtk_container_add(GTK_CONTAINER(file_menu), gtk_separator_menu_item_new());
+
 	dd->pref_menu_item = gtk_image_menu_item_new_from_stock("gtk-preferences", accel_group);
 	gtk_widget_add_accelerator(dd->pref_menu_item, "activate", accel_group,
 			GDK_p, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
@@ -525,11 +621,11 @@ void dict_gui_create_main_window(DictData *dd)
 	GtkWidget *main_box, *entry_box, *label_box;
 	GtkWidget *sep, *align, *scrolledwindow_results;
 	GdkPixbuf *icon;
-	GtkWidget *method_chooser, *radio, *label;
+	GtkWidget *method_chooser, *radio, *label, *button;
 
 	dd->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 	gtk_window_set_title(GTK_WINDOW(dd->window), _("Dictionary"));
-	gtk_window_set_default_size(GTK_WINDOW(dd->window), 500, 300);
+	gtk_window_set_default_size(GTK_WINDOW(dd->window), 580, 360);
 	gtk_widget_set_name(dd->window, "Xfce4Dict");
 
 	icon = gdk_pixbuf_new_from_inline(-1, dict_icon_data, FALSE, NULL);
@@ -543,14 +639,14 @@ void dict_gui_create_main_window(DictData *dd)
 	gtk_box_pack_start(GTK_BOX(main_box), create_file_menu(dd), FALSE, TRUE, 0);
 
 	/* entry box (label, entry, buttons) */
-	entry_box = gtk_hbox_new(FALSE, 10);
+	entry_box = gtk_hbox_new(FALSE, 5);
 	gtk_widget_show(entry_box);
 	gtk_container_set_border_width(GTK_CONTAINER(entry_box), 2);
 	gtk_box_pack_start(GTK_BOX(main_box), entry_box, FALSE, TRUE, 5);
 
 	label_box = gtk_hbox_new(FALSE, 5);
 	gtk_widget_show(label_box);
-	gtk_box_pack_start(GTK_BOX(entry_box), label_box, TRUE, TRUE, 0);
+	gtk_box_pack_start(GTK_BOX(entry_box), label_box, TRUE, TRUE, 5);
 
 	dd->main_combo = xfd_search_entry_new(_("Search term"));
 	gtk_widget_show(dd->main_combo);
@@ -570,16 +666,27 @@ void dict_gui_create_main_window(DictData *dd)
 	gtk_alignment_set_padding(GTK_ALIGNMENT(align), 0, 0, 10, 0);
 	gtk_widget_show(align);
 	gtk_container_add(GTK_CONTAINER(align), gtk_label_new(""));
-	gtk_box_pack_start(GTK_BOX(entry_box), align, FALSE, FALSE, 0);
+	gtk_box_pack_start(GTK_BOX(entry_box), align, FALSE, FALSE, 5);
 
-	dd->close_button = gtk_button_new_from_stock((dd->is_plugin) ? "gtk-close" : "gtk-quit");
-	gtk_widget_show(dd->close_button);
-	gtk_box_pack_end(GTK_BOX(entry_box), dd->close_button, FALSE, FALSE, 2);
-
-	/* insert it here and it will(hopefully) be placed before the Close button */
 	sep = gtk_vseparator_new();
 	gtk_widget_show(sep);
-	gtk_box_pack_end(GTK_BOX(entry_box), sep, FALSE, FALSE, 5);
+	gtk_box_pack_start(GTK_BOX(entry_box), sep, FALSE, FALSE, 2);
+
+	button = gtk_button_new_with_mnemonic(_("Speed _Reader"));
+	gtk_button_set_image(GTK_BUTTON(button),
+		gtk_image_new_from_stock(GTK_STOCK_JUSTIFY_CENTER, GTK_ICON_SIZE_MENU));
+	g_signal_connect(button, "clicked", G_CALLBACK(speedreader_clicked_cb), dd);
+	gtk_widget_show(button);
+	gtk_box_pack_start(GTK_BOX(entry_box), button, FALSE, FALSE, 2);
+
+	sep = gtk_vseparator_new();
+	gtk_widget_show(sep);
+	gtk_box_pack_start(GTK_BOX(entry_box), sep, FALSE, FALSE, 2);
+
+	dd->close_button = gtk_button_new_from_stock(
+		(dd->is_plugin) ? GTK_STOCK_CLOSE : GTK_STOCK_QUIT);
+	gtk_widget_show(dd->close_button);
+	gtk_box_pack_end(GTK_BOX(entry_box), dd->close_button, FALSE, FALSE, 0);
 
 	/* search method chooser */
 	method_chooser = gtk_hbox_new(FALSE, 0);
@@ -597,6 +704,8 @@ void dict_gui_create_main_window(DictData *dd)
 	gtk_box_pack_start(GTK_BOX(method_chooser), radio, FALSE, FALSE, 6);
 
 	radio = gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(radio), _("_Web Service"));
+	dd->radio_button_web = radio;
+	gtk_widget_set_sensitive(dd->radio_button_web, NZV(dd->web_url));
 	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), (dd->mode_in_use == DICTMODE_WEB));
 	g_signal_connect(radio, "toggled", G_CALLBACK(search_mode_web_toggled), dd);
 	gtk_widget_show(radio);
@@ -627,23 +736,30 @@ void dict_gui_create_main_window(DictData *dd)
 	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(dd->main_textview), GTK_WRAP_WORD);
 	dd->main_textbuffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(dd->main_textview));
 	gtk_text_buffer_create_tag(dd->main_textbuffer,
-			"bold",
+			TAG_HEADING,
 			"weight", PANGO_WEIGHT_BOLD,
-			"style", PANGO_STYLE_ITALIC,
-			"indent", 10,
 			"pixels-below-lines", 5, NULL);
 	gtk_text_buffer_create_tag(dd->main_textbuffer,
-			"error",
+			TAG_BOLD,
+			"weight", PANGO_WEIGHT_BOLD,
+			"indent", 10,
 			"style", PANGO_STYLE_ITALIC,
-			"foreground-gdk", &error_color, NULL);
+			"pixels-below-lines", 5, NULL);
+	dd->error_tag = gtk_text_buffer_create_tag(dd->main_textbuffer,
+			TAG_ERROR,
+			"style", PANGO_STYLE_ITALIC,
+			"foreground-gdk", dd->color_incorrect, NULL);
+	dd->success_tag = gtk_text_buffer_create_tag(dd->main_textbuffer,
+			TAG_SUCCESS,
+			"foreground-gdk", dd->color_correct, NULL);
 	dd->phon_tag = gtk_text_buffer_create_tag(dd->main_textbuffer,
-			"phonetic",
+			TAG_PHONETIC,
 			"style", PANGO_STYLE_ITALIC,
-			"foreground-gdk", dd->phon_color, NULL);
+			"foreground-gdk", dd->color_phonetic, NULL);
 	dd->link_tag = gtk_text_buffer_create_tag(dd->main_textbuffer,
-			"link",
+			TAG_LINK,
 			"underline", PANGO_UNDERLINE_SINGLE,
-			"foreground-gdk", dd->link_color, NULL);
+			"foreground-gdk", dd->color_link, NULL);
 
 	/* support for links (cross-references) for dictd responses */
 	{
@@ -670,6 +786,11 @@ void dict_gui_create_main_window(DictData *dd)
 		g_signal_connect(dd->main_textview, "populate-popup",
 			G_CALLBACK(textview_populate_popup_cb), dd);
 	}
+	/* tooltips */
+#if GTK_CHECK_VERSION(2, 12, 0)
+	gtk_widget_set_has_tooltip(dd->main_textview, TRUE);
+	g_signal_connect(dd->main_textview, "query-tooltip", G_CALLBACK(textview_query_tooltip_cb), dd);
+#endif
 
 	gtk_widget_show(dd->main_textview);
 	gtk_container_add(GTK_CONTAINER(scrolledwindow_results), dd->main_textview);
